;; Generated package description from paced.el
(define-package "paced" "1.1" "Predictive Abbreviation Completion and Expansion using Dictionaries" '((emacs "25.1") (async "1.9.1")) :url "https://savannah.nongnu.org/projects/paced-el/" :keywords '("convenience" "completion"))
